package java.awt.font;

import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.AffineTransform;
import java.awt.Shape;
import java.awt.geom.GeneralPath;
import sun.font.LayoutPathImpl;
import java.awt.Rectangle;
import sun.font.GraphicComponent;
import sun.font.AttributeValues;
import sun.font.CoreMetrics;
import sun.text.CodePointIterator;
import sun.font.FontResolver;
import java.text.AttributedCharacterIterator;
import java.util.Map;
import java.text.AttributedString;
import java.awt.Font;
import java.awt.geom.Rectangle2D;

public final class TextLayout implements Cloneable
{
    private int characterCount;
    private boolean isVerticalLine;
    private byte baseline;
    private float[] baselineOffsets;
    private TextLine textLine;
    private TextLine.TextLineMetrics lineMetrics;
    private float visibleAdvance;
    private int hashCodeCache;
    private boolean cacheIsValid;
    private float justifyRatio;
    private static final float ALREADY_JUSTIFIED = -53.9f;
    private static float dx;
    private static float dy;
    private Rectangle2D naturalBounds;
    private Rectangle2D boundsRect;
    private boolean caretsInLigaturesAreAllowed;
    public static final TextLayout.CaretPolicy DEFAULT_CARET_POLICY;
    
    public TextLayout(String s, Font value, FontRenderContext fontRenderContext) {
        this.isVerticalLine = false;
        this.lineMetrics = null;
        this.cacheIsValid = false;
        this.naturalBounds = null;
        this.boundsRect = null;
        this.caretsInLigaturesAreAllowed = false;
        if (value == null) {
            throw new IllegalArgumentException("Null font passed to TextLayout constructor.");
        }
        if (s == null) {
            throw new IllegalArgumentException("Null string passed to TextLayout constructor.");
        }
        if (s.length() == 0) {
            throw new IllegalArgumentException("Zero length string passed to TextLayout constructor.");
        }
        Map<TextAttribute, Object> attributes = null;
        if (value.hasLayoutAttributes()) {
            attributes = value.getAttributes();
        }
        char[] charArray = s.toCharArray();
        if (sameBaselineUpTo(value, charArray, 0, charArray.length) == charArray.length) {
            this.fastInit(charArray, value, attributes, fontRenderContext);
        }
        else {
            Object o = (attributes == null) ? new AttributedString(s) : new AttributedString(s, attributes);
            o.addAttribute(TextAttribute.FONT, value);
            this.standardInit(o.getIterator(), charArray, fontRenderContext);
        }
        return;
    }
    
    public TextLayout(String text, Map<? extends AttributedCharacterIterator.Attribute, ?> attributes, FontRenderContext fontRenderContext) {
        this.isVerticalLine = false;
        this.lineMetrics = null;
        this.cacheIsValid = false;
        this.naturalBounds = null;
        this.boundsRect = null;
        this.caretsInLigaturesAreAllowed = false;
        if (text == null) {
            throw new IllegalArgumentException("Null string passed to TextLayout constructor.");
        }
        if (attributes == null) {
            throw new IllegalArgumentException("Null map passed to TextLayout constructor.");
        }
        if (text.length() == 0) {
            throw new IllegalArgumentException("Zero length string passed to TextLayout constructor.");
        }
        char[] charArray = text.toCharArray();
        Font singleFont = singleFont(charArray, 0, charArray.length, attributes);
        if (singleFont != null) {
            this.fastInit(charArray, singleFont, attributes, fontRenderContext);
        }
        else {
            this.standardInit(new AttributedString(text, attributes).getIterator(), charArray, fontRenderContext);
        }
        return;
    }
    
    private static Font singleFont(char[] text, int start, int limit, Map<? extends AttributedCharacterIterator.Attribute, ?> map) {
        if (map.get(TextAttribute.CHAR_REPLACEMENT) != null) {
            return null;
        }
        Font font = null;
        try {
            font = (Font)map.get(TextAttribute.FONT);
        }
        catch (ClassCastException ex) {}
        if (font == null) {
            if (map.get(TextAttribute.FAMILY) != null) {
                font = Font.getFont(map);
                if (font.canDisplayUpTo(text, start, limit) != -1) {
                    return null;
                }
            }
            else {
                FontResolver instance = FontResolver.getInstance();
                CodePointIterator create = CodePointIterator.create(text, start, limit);
                int nextFontRunIndex = instance.nextFontRunIndex(create);
                if (create.charIndex() == limit) {
                    font = instance.getFont(nextFontRunIndex, map);
                }
            }
        }
        if (sameBaselineUpTo(font, text, start, limit) != limit) {
            return null;
        }
        return font;
    }
    
    public TextLayout(AttributedCharacterIterator attributedCharacterIterator, FontRenderContext fontRenderContext) {
        this.isVerticalLine = false;
        this.lineMetrics = null;
        this.cacheIsValid = false;
        this.naturalBounds = null;
        this.boundsRect = null;
        this.caretsInLigaturesAreAllowed = false;
        if (attributedCharacterIterator == null) {
            throw new IllegalArgumentException("Null iterator passed to TextLayout constructor.");
        }
        int beginIndex = attributedCharacterIterator.getBeginIndex();
        int endIndex = attributedCharacterIterator.getEndIndex();
        if (beginIndex == endIndex) {
            throw new IllegalArgumentException("Zero length iterator passed to TextLayout constructor.");
        }
        int n = endIndex - beginIndex;
        attributedCharacterIterator.first();
        char[] array = new char[n];
        int n2 = 0;
        for (int i = attributedCharacterIterator.first(); i != '\uffff'; i = attributedCharacterIterator.next()) {
            array[n2++] = i;
        }
        attributedCharacterIterator.first();
        if (attributedCharacterIterator.getRunLimit() == endIndex) {
            Map<AttributedCharacterIterator.Attribute, Object> attributes = attributedCharacterIterator.getAttributes();
            Font singleFont = singleFont(array, 0, n, attributes);
            if (singleFont != null) {
                this.fastInit(array, singleFont, attributes, fontRenderContext);
                return;
            }
        }
        this.standardInit(attributedCharacterIterator, array, fontRenderContext);
        return;
    }
    
    TextLayout(TextLine textLine, byte baseline, float[] baselineOffsets, float justifyRatio) {
        this.isVerticalLine = false;
        this.lineMetrics = null;
        this.cacheIsValid = false;
        this.naturalBounds = null;
        this.boundsRect = null;
        this.caretsInLigaturesAreAllowed = false;
        this.characterCount = textLine.characterCount();
        this.baseline = baseline;
        this.baselineOffsets = baselineOffsets;
        this.textLine = textLine;
        this.justifyRatio = justifyRatio;
    }
    
    private void paragraphInit(byte baseline, CoreMetrics coreMetrics, Map<? extends AttributedCharacterIterator.Attribute, ?> map, char[] array) {
        this.baseline = baseline;
        this.baselineOffsets = TextLine.getNormalizedOffsets(coreMetrics.baselineOffsets, this.baseline);
        this.justifyRatio = AttributeValues.getJustification(map);
        NumericShaper numericShaping = AttributeValues.getNumericShaping(map);
        if (numericShaping != null) {
            numericShaping.shape(array, 0, array.length);
        }
    }
    
    private void fastInit(char[] chars, Font font, Map<? extends AttributedCharacterIterator.Attribute, ?> map, FontRenderContext frc) {
        this.isVerticalLine = false;
        CoreMetrics value = CoreMetrics.get(font.getLineMetrics(chars, 0, chars.length, frc));
        int baseline = (byte)value.baselineIndex;
        if (map == null) {
            this.baseline = baseline;
            this.baselineOffsets = value.baselineOffsets;
            this.justifyRatio = 1.0f;
        }
        else {
            this.paragraphInit((byte)baseline, value, map, chars);
        }
        this.characterCount = chars.length;
        this.textLine = TextLine.fastCreateTextLine(frc, chars, font, value, map);
    }
    
    private void standardInit(AttributedCharacterIterator attributedCharacterIterator, char[] chars, FontRenderContext frc) {
        this.characterCount = chars.length;
        Map<AttributedCharacterIterator.Attribute, Object> attributes = attributedCharacterIterator.getAttributes();
        if (TextLine.advanceToFirstFont(attributedCharacterIterator)) {
            Font fontAtCurrentPos = TextLine.getFontAtCurrentPos(attributedCharacterIterator);
            int beginIndex = attributedCharacterIterator.getIndex() - attributedCharacterIterator.getBeginIndex();
            CoreMetrics value = CoreMetrics.get(fontAtCurrentPos.getLineMetrics(chars, beginIndex, beginIndex + 1, frc));
            this.paragraphInit((byte)value.baselineIndex, value, attributes, chars);
        }
        else {
            GraphicAttribute graphic = (GraphicAttribute)attributes.get(TextAttribute.CHAR_REPLACEMENT);
            this.paragraphInit(getBaselineFromGraphic(graphic), GraphicComponent.createCoreMetrics(graphic), attributes, chars);
        }
        this.textLine = TextLine.standardCreateTextLine(frc, attributedCharacterIterator, chars, this.baselineOffsets);
    }
    
    private void ensureCache() {
        if (!this.cacheIsValid) {
            this.buildCache();
        }
    }
    
    private void buildCache() {
        this.lineMetrics = this.textLine.getMetrics();
        if (this.textLine.isDirectionLTR()) {
            int n2;
            for (n2 = this.characterCount - 1; n2 != -1 && this.textLine.isCharSpace(this.textLine.visualToLogical(n2)); --n2) {}
            if (n2 == this.characterCount - 1) {
                this.visibleAdvance = this.lineMetrics.advance;
            }
            else if (n2 == -1) {
                this.visibleAdvance = 0.0f;
            }
            else {
                int visualToLogical = this.textLine.visualToLogical(n2);
                this.visibleAdvance = this.textLine.getCharLinePosition(visualToLogical) + this.textLine.getCharAdvance(visualToLogical);
            }
        }
        else {
            int n;
            for (n = 0; n != this.characterCount && this.textLine.isCharSpace(this.textLine.visualToLogical(n)); ++n) {}
            if (n == this.characterCount) {
                this.visibleAdvance = 0.0f;
            }
            else if (n == 0) {
                this.visibleAdvance = this.lineMetrics.advance;
            }
            else {
                this.visibleAdvance = this.lineMetrics.advance - this.textLine.getCharLinePosition(this.textLine.visualToLogical(n));
            }
        }
        this.naturalBounds = null;
        this.boundsRect = null;
        this.hashCodeCache = 0;
        this.cacheIsValid = true;
    }
    
    private Rectangle2D getNaturalBounds() {
        this.ensureCache();
        if (this.naturalBounds == null) {
            this.naturalBounds = this.textLine.getItalicBounds();
        }
        return this.naturalBounds;
    }
    
    @Override
    protected Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException cause) {
            throw new InternalError(cause);
        }
    }
    
    private void checkTextHit(TextHitInfo textHitInfo) {
        if (textHitInfo == null) {
            throw new IllegalArgumentException("TextHitInfo is null.");
        }
        if (textHitInfo.getInsertionIndex() < 0 || textHitInfo.getInsertionIndex() > this.characterCount) {
            throw new IllegalArgumentException("TextHitInfo is out of range");
        }
        return;
    }
    
    public TextLayout getJustifiedLayout(float n) {
        if (n <= 0.0f) {
            throw new IllegalArgumentException("justificationWidth <= 0 passed to TextLayout.getJustifiedLayout()");
        }
        if (this.justifyRatio == -53.9f) {
            throw new Error("Can't justify again.");
        }
        this.ensureCache();
        int characterCount;
        for (characterCount = this.characterCount; characterCount > 0 && this.textLine.isCharWhitespace(characterCount - 1); --characterCount) {}
        TextLine justifiedLine = this.textLine.getJustifiedLine(n, this.justifyRatio, 0, characterCount);
        if (justifiedLine != null) {
            return new TextLayout(justifiedLine, this.baseline, this.baselineOffsets, -53.9f);
        }
        return this;
    }
    
    protected void handleJustify(float n) {
    }
    
    public byte getBaseline() {
        return this.baseline;
    }
    
    public float[] getBaselineOffsets() {
        float[] array = new float[this.baselineOffsets.length];
        System.arraycopy(this.baselineOffsets, 0, array, 0, array.length);
        return array;
    }
    
    public float getAdvance() {
        this.ensureCache();
        return this.lineMetrics.advance;
    }
    
    public float getVisibleAdvance() {
        this.ensureCache();
        return this.visibleAdvance;
    }
    
    public float getAscent() {
        this.ensureCache();
        return this.lineMetrics.ascent;
    }
    
    public float getDescent() {
        this.ensureCache();
        return this.lineMetrics.descent;
    }
    
    public float getLeading() {
        this.ensureCache();
        return this.lineMetrics.leading;
    }
    
    public Rectangle2D getBounds() {
        this.ensureCache();
        if (this.boundsRect == null) {
            Rectangle2D visualBounds = this.textLine.getVisualBounds();
            if (TextLayout.dx != 0.0f || TextLayout.dy != 0.0f) {
                visualBounds.setRect(visualBounds.getX() - (double)TextLayout.dx, visualBounds.getY() - (double)TextLayout.dy, visualBounds.getWidth(), visualBounds.getHeight());
            }
            this.boundsRect = visualBounds;
        }
        Rectangle2D.Float o = new Rectangle2D.Float();
        o.setRect(this.boundsRect);
        return o;
    }
    
    public Rectangle getPixelBounds(FontRenderContext fontRenderContext, float n, float n2) {
        return this.textLine.getPixelBounds(fontRenderContext, n, n2);
    }
    
    public boolean isLeftToRight() {
        return this.textLine.isDirectionLTR();
    }
    
    public boolean isVertical() {
        return this.isVerticalLine;
    }
    
    public int getCharacterCount() {
        return this.characterCount;
    }
    
    private float[] getCaretInfo(int n, Rectangle2D rectangle2D, float[] array) {
        float n4;
        float n3;
        float n7;
        float n6;
        if (n == 0 || n == this.characterCount) {
            int n9;
            float charLinePosition2;
            if (n == this.characterCount) {
                n9 = this.textLine.visualToLogical(this.characterCount - 1);
                charLinePosition2 = this.textLine.getCharLinePosition(n9) + this.textLine.getCharAdvance(n9);
            }
            else {
                n9 = this.textLine.visualToLogical(n);
                charLinePosition2 = this.textLine.getCharLinePosition(n9);
            }
            float charAngle3 = this.textLine.getCharAngle(n9);
            float n10 = charLinePosition2 + charAngle3 * this.textLine.getCharShift(n9);
            n7 = (n4 = n10 + charAngle3 * this.textLine.getCharAscent(n9));
            n6 = (n3 = n10 - charAngle3 * this.textLine.getCharDescent(n9));
        }
        else {
            int visualToLogical = this.textLine.visualToLogical(n - 1);
            float charAngle = this.textLine.getCharAngle(visualToLogical);
            float n2 = this.textLine.getCharLinePosition(visualToLogical) + this.textLine.getCharAdvance(visualToLogical);
            if (charAngle != 0.0f) {
                float n5 = n2 + charAngle * this.textLine.getCharShift(visualToLogical);
                n4 = n5 + charAngle * this.textLine.getCharAscent(visualToLogical);
                n3 = n5 - charAngle * this.textLine.getCharDescent(visualToLogical);
            }
            else {
                n3 = (n4 = n2);
            }
            int visualToLogical2 = this.textLine.visualToLogical(n);
            float charAngle2 = this.textLine.getCharAngle(visualToLogical2);
            float charLinePosition = this.textLine.getCharLinePosition(visualToLogical2);
            if (charAngle2 != 0.0f) {
                float n8 = charLinePosition + charAngle2 * this.textLine.getCharShift(visualToLogical2);
                n7 = n8 + charAngle2 * this.textLine.getCharAscent(visualToLogical2);
                n6 = n8 - charAngle2 * this.textLine.getCharDescent(visualToLogical2);
            }
            else {
                n6 = (n7 = charLinePosition);
            }
        }
        float n11 = (n4 + n7) / 2.0f;
        float n12 = (n3 + n6) / 2.0f;
        if (array == null) {
            array = new float[2];
        }
        if (this.isVerticalLine) {
            array[1] = (float)((double)(n11 - n12) / rectangle2D.getWidth());
            array[0] = (float)((double)n11 + (double)array[1] * rectangle2D.getX());
            return array;
        }
        array[1] = (float)((double)(n11 - n12) / rectangle2D.getHeight());
        array[0] = (float)((double)n12 + (double)array[1] * rectangle2D.getMaxY());
        return array;
    }
    
    public float[] getCaretInfo(TextHitInfo textHitInfo, Rectangle2D rectangle2D) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        return this.getCaretInfoTestInternal(textHitInfo, rectangle2D);
    }
    
    private float[] getCaretInfoTestInternal(TextHitInfo textHitInfo, Rectangle2D rectangle2D) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        float[] array = new float[6];
        this.getCaretInfo(this.hitToCaret(textHitInfo), rectangle2D, array);
        int charIndex = textHitInfo.getCharIndex();
        int leadingEdge = textHitInfo.isLeadingEdge();
        int directionLTR = this.textLine.isDirectionLTR();
        int n = !this.isVertical();
        double n6;
        double n7;
        double n8;
        double n9;
        if (charIndex == -1 || charIndex == this.characterCount) {
            TextLine.TextLineMetrics metrics2 = this.textLine.getMetrics();
            int n11 = directionLTR == ((charIndex == -1) ? 1 : 0);
            if (!(n == 0)) {
                n9 = (n8 = (n11 ? 0.0 : ((double)metrics2.advance)));
                n6 = (double)(-metrics2.ascent);
                n7 = (double)metrics2.descent;
            }
            else {
                n7 = (n6 = (n11 ? 0.0 : ((double)metrics2.advance)));
                n8 = (double)metrics2.descent;
                n9 = (double)metrics2.ascent;
            }
        }
        else {
            CoreMetrics coreMetrics = this.textLine.getCoreMetricsAt(charIndex);
            double n2 = (double)coreMetrics.italicAngle;
            double n3 = (double)this.textLine.getCharLinePosition(charIndex, (boolean)leadingEdge);
            if (coreMetrics.baselineIndex < 0) {
                TextLine.TextLineMetrics metrics = this.textLine.getMetrics();
                if (!(n == 0)) {
                    n9 = (n8 = n3);
                    if (coreMetrics.baselineIndex == -1) {
                        n6 = (double)(-metrics.ascent);
                        n7 = n6 + (double)coreMetrics.height;
                    }
                    else {
                        n7 = (double)metrics.descent;
                        n6 = n7 - (double)coreMetrics.height;
                    }
                }
                else {
                    n7 = (n6 = n3);
                    n8 = (double)metrics.descent;
                    n9 = (double)metrics.ascent;
                }
            }
            else {
                float n4 = this.baselineOffsets[coreMetrics.baselineIndex];
                if (!(n == 0)) {
                    double n10 = n3 + n2 * (double)coreMetrics.ssOffset;
                    n8 = n10 + n2 * (double)coreMetrics.ascent;
                    n9 = n10 - n2 * (double)coreMetrics.descent;
                    n6 = (double)(n4 - coreMetrics.ascent);
                    n7 = (double)(n4 + coreMetrics.descent);
                }
                else {
                    double n5 = n3 - n2 * (double)coreMetrics.ssOffset;
                    n6 = n5 + n2 * (double)coreMetrics.ascent;
                    n7 = n5 - n2 * (double)coreMetrics.descent;
                    n8 = (double)(n4 + coreMetrics.ascent);
                    n9 = (double)(n4 + coreMetrics.descent);
                }
            }
        }
        array[2] = (float)n8;
        array[3] = (float)n6;
        array[4] = (float)n9;
        array[5] = (float)n7;
        return array;
    }
    
    public float[] getCaretInfo(TextHitInfo textHitInfo) {
        return this.getCaretInfo(textHitInfo, this.getNaturalBounds());
    }
    
    private int hitToCaret(TextHitInfo textHitInfo) {
        int charIndex = textHitInfo.getCharIndex();
        if (charIndex < 0) {
            return this.textLine.isDirectionLTR() ? 0 : this.characterCount;
        }
        if (charIndex >= this.characterCount) {
            return this.textLine.isDirectionLTR() ? this.characterCount : 0;
        }
        int logicalToVisual = this.textLine.logicalToVisual(charIndex);
        if (textHitInfo.isLeadingEdge() != this.textLine.isCharLTR(charIndex)) {
            ++logicalToVisual;
        }
        return logicalToVisual;
    }
    
    private TextHitInfo caretToHit(int n) {
        if (n == 0 || n == this.characterCount) {
            if (n == this.characterCount == this.textLine.isDirectionLTR()) {
                return TextHitInfo.leading(this.characterCount);
            }
            return TextHitInfo.trailing(-1);
        }
        else {
            int visualToLogical = this.textLine.visualToLogical(n);
            return this.textLine.isCharLTR(visualToLogical) ? TextHitInfo.leading(visualToLogical) : TextHitInfo.trailing(visualToLogical);
        }
    }
    
    private boolean caretIsValid(int n) {
        if (n == this.characterCount || n == 0) {
            return true;
        }
        int n2 = this.textLine.visualToLogical(n);
        if (!this.textLine.isCharLTR(n2)) {
            n2 = this.textLine.visualToLogical(n - 1);
            if (this.textLine.isCharLTR(n2)) {
                return true;
            }
        }
        return this.textLine.caretAtOffsetIsValid(n2);
    }
    
    public TextHitInfo getNextRightHit(TextHitInfo textHitInfo) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        int hitToCaret = this.hitToCaret(textHitInfo);
        if (hitToCaret == this.characterCount) {
            return null;
        }
        do {
            ++hitToCaret;
        } while (!this.caretIsValid(hitToCaret));
        return this.caretToHit(hitToCaret);
    }
    
    public TextHitInfo getNextRightHit(int n, TextLayout.CaretPolicy caretPolicy) {
        if (n < 0 || n > this.characterCount) {
            throw new IllegalArgumentException("Offset out of bounds in TextLayout.getNextRightHit()");
        }
        if (caretPolicy == null) {
            throw new IllegalArgumentException("Null CaretPolicy passed to TextLayout.getNextRightHit()");
        }
        TextHitInfo afterOffset = TextHitInfo.afterOffset(n);
        TextHitInfo nextRightHit = this.getNextRightHit(caretPolicy.getStrongCaret(afterOffset, afterOffset.getOtherHit(), this));
        if (nextRightHit != null) {
            return caretPolicy.getStrongCaret(this.getVisualOtherHit(nextRightHit), nextRightHit, this);
        }
        return null;
    }
    
    public TextHitInfo getNextRightHit(int n) {
        return this.getNextRightHit(n, TextLayout.DEFAULT_CARET_POLICY);
    }
    
    public TextHitInfo getNextLeftHit(TextHitInfo textHitInfo) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        int hitToCaret = this.hitToCaret(textHitInfo);
        if (hitToCaret == 0) {
            return null;
        }
        do {
            --hitToCaret;
        } while (!this.caretIsValid(hitToCaret));
        return this.caretToHit(hitToCaret);
    }
    
    public TextHitInfo getNextLeftHit(int n, TextLayout.CaretPolicy caretPolicy) {
        if (caretPolicy == null) {
            throw new IllegalArgumentException("Null CaretPolicy passed to TextLayout.getNextLeftHit()");
        }
        if (n < 0 || n > this.characterCount) {
            throw new IllegalArgumentException("Offset out of bounds in TextLayout.getNextLeftHit()");
        }
        TextHitInfo afterOffset = TextHitInfo.afterOffset(n);
        TextHitInfo nextLeftHit = this.getNextLeftHit(caretPolicy.getStrongCaret(afterOffset, afterOffset.getOtherHit(), this));
        if (nextLeftHit != null) {
            return caretPolicy.getStrongCaret(this.getVisualOtherHit(nextLeftHit), nextLeftHit, this);
        }
        return null;
    }
    
    public TextHitInfo getNextLeftHit(int n) {
        return this.getNextLeftHit(n, TextLayout.DEFAULT_CARET_POLICY);
    }
    
    public TextHitInfo getVisualOtherHit(TextHitInfo textHitInfo) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        int charIndex = textHitInfo.getCharIndex();
        int n2;
        int charLTR;
        if (charIndex == -1 || charIndex == this.characterCount) {
            int n3;
            if (this.textLine.isDirectionLTR() == (charIndex == -1)) {
                n3 = 0;
            }
            else {
                n3 = this.characterCount - 1;
            }
            n2 = this.textLine.visualToLogical(n3);
            if (this.textLine.isDirectionLTR() == (charIndex == -1)) {
                charLTR = this.textLine.isCharLTR(n2);
            }
            else {
                charLTR = !this.textLine.isCharLTR(n2);
            }
        }
        else {
            int logicalToVisual = this.textLine.logicalToVisual(charIndex);
            int n;
            if (this.textLine.isCharLTR(charIndex) == textHitInfo.isLeadingEdge()) {
                --logicalToVisual;
                n = false;
            }
            else {
                ++logicalToVisual;
                n = true;
            }
            if (logicalToVisual > -1 && logicalToVisual < this.characterCount) {
                n2 = this.textLine.visualToLogical(logicalToVisual);
                charLTR = (n == (this.textLine.isCharLTR(n2) ? 1 : 0));
            }
            else {
                n2 = ((n == (this.textLine.isDirectionLTR() ? 1 : 0)) ? this.characterCount : -1);
                charLTR = (n2 == this.characterCount);
            }
        }
        return charLTR ? TextHitInfo.leading(n2) : TextHitInfo.trailing(n2);
    }
    
    private double[] getCaretPath(TextHitInfo textHitInfo, Rectangle2D rectangle2D) {
        float[] caretInfo = this.getCaretInfo(textHitInfo, rectangle2D);
        return new double[] { (double)caretInfo[2], (double)caretInfo[3], (double)caretInfo[4], (double)caretInfo[5] };
    }
    
    private double[] getCaretPath(int n, Rectangle2D rectangle2D, boolean b) {
        float[] caretInfo = this.getCaretInfo(n, rectangle2D, (float[])null);
        double n2 = (double)caretInfo[0];
        double n3 = (double)caretInfo[1];
        double n4 = -3141.59;
        double n5 = -2.7;
        double x = rectangle2D.getX();
        double n6 = x + rectangle2D.getWidth();
        double y = rectangle2D.getY();
        double n7 = y + rectangle2D.getHeight();
        boolean n8 = false;
        double n9;
        double n10;
        double n11;
        double n12;
        if (this.isVerticalLine) {
            if (n3 >= 0.0) {
                n11 = x;
                n12 = n6;
            }
            else {
                n12 = x;
                n11 = n6;
            }
            n10 = n2 + n11 * n3;
            n9 = n2 + n12 * n3;
            if (b) {
                if (n10 < y) {
                    if (n3 <= 0.0 || n9 <= y) {
                        n9 = (n10 = y);
                    }
                    else {
                        n8 = true;
                        n10 = y;
                        n5 = y;
                        n4 = n12 + (y - n9) / n3;
                        if (n9 > n7) {
                            n9 = n7;
                        }
                    }
                }
                else if (n9 > n7) {
                    if (n3 >= 0.0 || n10 >= n7) {
                        n9 = (n10 = n7);
                    }
                    else {
                        n8 = true;
                        n9 = n7;
                        n5 = n7;
                        n4 = n11 + (n7 - n12) / n3;
                    }
                }
            }
        }
        else {
            if (n3 >= 0.0) {
                n10 = n7;
                n9 = y;
            }
            else {
                n9 = n7;
                n10 = y;
            }
            n11 = n2 - n10 * n3;
            n12 = n2 - n9 * n3;
            if (b) {
                if (n11 < x) {
                    if (n3 <= 0.0 || n12 <= x) {
                        n12 = (n11 = x);
                    }
                    else {
                        n8 = true;
                        n11 = x;
                        n4 = x;
                        n5 = n9 - (x - n12) / n3;
                        if (n12 > n6) {
                            n12 = n6;
                        }
                    }
                }
                else if (n12 > n6) {
                    if (n3 >= 0.0 || n11 >= n6) {
                        n12 = (n11 = n6);
                    }
                    else {
                        n8 = true;
                        n12 = n6;
                        n4 = n6;
                        n5 = n10 - (n6 - n11) / n3;
                    }
                }
            }
        }
        return n8 ? new double[] { n11, n10, n4, n5, n12, n9 } : new double[] { n11, n10, n12, n9 };
    }
    
    private static GeneralPath pathToShape(double[] array, boolean b, LayoutPathImpl layoutPathImpl) {
        GeneralPath generalPath = new GeneralPath(0, array.length);
        generalPath.moveTo((float)array[0], (float)array[1]);
        for (int i = 2; i < array.length; i += 2) {
            generalPath.lineTo((float)array[i], (float)array[i + 1]);
        }
        if (b) {
            generalPath.closePath();
        }
        if (layoutPathImpl != null) {
            generalPath = (GeneralPath)layoutPathImpl.mapShape(generalPath);
            return generalPath;
        }
        return generalPath;
    }
    
    public Shape getCaretShape(TextHitInfo textHitInfo, Rectangle2D rectangle2D) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        if (rectangle2D == null) {
            throw new IllegalArgumentException("Null Rectangle2D passed to TextLayout.getCaret()");
        }
        return pathToShape(this.getCaretPath(textHitInfo, rectangle2D), false, this.textLine.getLayoutPath());
    }
    
    public Shape getCaretShape(TextHitInfo textHitInfo) {
        return this.getCaretShape(textHitInfo, this.getNaturalBounds());
    }
    
    private final TextHitInfo getStrongHit(TextHitInfo textHitInfo, TextHitInfo textHitInfo2) {
        int characterLevel = this.getCharacterLevel(textHitInfo.getCharIndex());
        int characterLevel2 = this.getCharacterLevel(textHitInfo2.getCharIndex());
        if (characterLevel == characterLevel2) {
            if (textHitInfo2.isLeadingEdge() && !textHitInfo.isLeadingEdge()) {
                return textHitInfo2;
            }
            return textHitInfo;
        }
        else {
            return (characterLevel < characterLevel2) ? textHitInfo : textHitInfo2;
        }
    }
    
    public byte getCharacterLevel(int n) {
        if (n < -1 || n > this.characterCount) {
            throw new IllegalArgumentException("Index is out of range in getCharacterLevel.");
        }
        this.ensureCache();
        if (n == -1 || n == this.characterCount) {
            return (byte)!this.textLine.isDirectionLTR();
        }
        return this.textLine.getCharLevel(n);
    }
    
    public Shape[] getCaretShapes(int n, Rectangle2D rectangle2D, TextLayout.CaretPolicy caretPolicy) {
        this.ensureCache();
        if (n < 0 || n > this.characterCount) {
            throw new IllegalArgumentException("Offset out of bounds in TextLayout.getCaretShapes()");
        }
        if (rectangle2D == null) {
            throw new IllegalArgumentException("Null Rectangle2D passed to TextLayout.getCaretShapes()");
        }
        if (caretPolicy == null) {
            throw new IllegalArgumentException("Null CaretPolicy passed to TextLayout.getCaretShapes()");
        }
        Shape[] array = new Shape[2];
        TextHitInfo afterOffset = TextHitInfo.afterOffset(n);
        int hitToCaret = this.hitToCaret(afterOffset);
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        GeneralPath pathToShape = pathToShape(this.getCaretPath(afterOffset, rectangle2D), false, layoutPath);
        TextHitInfo otherHit = afterOffset.getOtherHit();
        if (hitToCaret == this.hitToCaret(otherHit)) {
            array[0] = pathToShape;
        }
        else {
            GeneralPath pathToShape2 = pathToShape(this.getCaretPath(otherHit, rectangle2D), false, layoutPath);
            if (caretPolicy.getStrongCaret(afterOffset, otherHit, this).equals(afterOffset)) {
                array[0] = pathToShape;
                array[1] = pathToShape2;
            }
            else {
                array[0] = pathToShape2;
                array[1] = pathToShape;
            }
        }
        return array;
    }
    
    public Shape[] getCaretShapes(int n, Rectangle2D rectangle2D) {
        return this.getCaretShapes(n, rectangle2D, TextLayout.DEFAULT_CARET_POLICY);
    }
    
    public Shape[] getCaretShapes(int n) {
        return this.getCaretShapes(n, this.getNaturalBounds(), TextLayout.DEFAULT_CARET_POLICY);
    }
    
    private GeneralPath boundingShape(double[] array, double[] array2) {
        GeneralPath pathToShape = pathToShape(array, false, null);
        int n;
        if (this.isVerticalLine) {
            n = (array[1] > array[array.length - 1] == array2[1] > array2[array2.length - 1]);
        }
        else {
            n = (array[0] > array[array.length - 2] == array2[0] > array2[array2.length - 2]);
        }
        int i;
        int length;
        int n2;
        if (n != 0) {
            i = array2.length - 2;
            length = -2;
            n2 = -2;
        }
        else {
            i = 0;
            length = array2.length;
            n2 = 2;
        }
        for (; i != length; i += n2) {
            pathToShape.lineTo((float)array2[i], (float)array2[i + 1]);
        }
        pathToShape.closePath();
        return pathToShape;
    }
    
    private GeneralPath caretBoundingShape(int n, int n2, Rectangle2D rectangle2D) {
        if (n > n2) {
            n = (n2 = n2);
        }
        return this.boundingShape(this.getCaretPath(n, rectangle2D, true), this.getCaretPath(n2, rectangle2D, true));
    }
    
    private GeneralPath leftShape(Rectangle2D rectangle2D) {
        double[] array;
        if (this.isVerticalLine) {
            array = new double[] { rectangle2D.getX(), rectangle2D.getY(), rectangle2D.getX() + rectangle2D.getWidth(), rectangle2D.getY() };
        }
        else {
            array = new double[] { rectangle2D.getX(), rectangle2D.getY() + rectangle2D.getHeight(), rectangle2D.getX(), rectangle2D.getY() };
        }
        return this.boundingShape(array, this.getCaretPath(0, rectangle2D, true));
    }
    
    private GeneralPath rightShape(Rectangle2D rectangle2D) {
        double[] array;
        if (this.isVerticalLine) {
            array = new double[] { rectangle2D.getX(), rectangle2D.getY() + rectangle2D.getHeight(), rectangle2D.getX() + rectangle2D.getWidth(), rectangle2D.getY() + rectangle2D.getHeight() };
        }
        else {
            array = new double[] { rectangle2D.getX() + rectangle2D.getWidth(), rectangle2D.getY() + rectangle2D.getHeight(), rectangle2D.getX() + rectangle2D.getWidth(), rectangle2D.getY() };
        }
        return this.boundingShape(this.getCaretPath(this.characterCount, rectangle2D, true), array);
    }
    
    public int[] getLogicalRangesForVisualSelection(TextHitInfo textHitInfo, TextHitInfo textHitInfo2) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        this.checkTextHit(textHitInfo2);
        boolean[] array = new boolean[this.characterCount];
        int i = this.hitToCaret(textHitInfo);
        int hitToCaret = this.hitToCaret(textHitInfo2);
        if (i > hitToCaret) {
            i = (hitToCaret = hitToCaret);
        }
        if (i < hitToCaret) {
            for (; i < hitToCaret; ++i) {
                array[this.textLine.visualToLogical(i)] = true;
            }
        }
        int n = 0;
        boolean n2 = false;
        for (int j = 0; j < this.characterCount; ++j) {
            if ((array[j] ? 1 : 0) != n2) {
                n2 = (n2 == 0);
                if (!(n2 == 0)) {
                    ++n;
                }
            }
        }
        int[] array2 = new int[n * 2];
        int n3 = 0;
        boolean n4 = false;
        for (int k = 0; k < this.characterCount; ++k) {
            if ((array[k] ? 1 : 0) != n4) {
                array2[n3++] = k;
                n4 = (n4 == 0);
            }
        }
        if (!(n4 == 0)) {
            array2[n3++] = this.characterCount;
            return array2;
        }
        return array2;
    }
    
    public Shape getVisualHighlightShape(TextHitInfo textHitInfo, TextHitInfo textHitInfo2, Rectangle2D rectangle2D) {
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        this.checkTextHit(textHitInfo2);
        if (rectangle2D == null) {
            throw new IllegalArgumentException("Null Rectangle2D passed to TextLayout.getVisualHighlightShape()");
        }
        GeneralPath generalPath = new GeneralPath(0);
        int hitToCaret = this.hitToCaret(textHitInfo);
        int hitToCaret2 = this.hitToCaret(textHitInfo2);
        generalPath.append(this.caretBoundingShape(hitToCaret, hitToCaret2, rectangle2D), false);
        if (hitToCaret == 0 || hitToCaret2 == 0) {
            GeneralPath leftShape = this.leftShape(rectangle2D);
            if (!leftShape.getBounds().isEmpty()) {
                generalPath.append(leftShape, false);
            }
        }
        if (hitToCaret == this.characterCount || hitToCaret2 == this.characterCount) {
            GeneralPath rightShape = this.rightShape(rectangle2D);
            if (!rightShape.getBounds().isEmpty()) {
                generalPath.append(rightShape, false);
            }
        }
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            generalPath = (GeneralPath)layoutPath.mapShape(generalPath);
            return generalPath;
        }
        return generalPath;
    }
    
    public Shape getVisualHighlightShape(TextHitInfo textHitInfo, TextHitInfo textHitInfo2) {
        return this.getVisualHighlightShape(textHitInfo, textHitInfo2, this.getNaturalBounds());
    }
    
    public Shape getLogicalHighlightShape(int i, int n, Rectangle2D rectangle2D) {
        if (rectangle2D == null) {
            throw new IllegalArgumentException("Null Rectangle2D passed to TextLayout.getLogicalHighlightShape()");
        }
        this.ensureCache();
        if (i > n) {
            i = (n = n);
        }
        if (i < 0 || n > this.characterCount) {
            throw new IllegalArgumentException("Range is invalid in TextLayout.getLogicalHighlightShape()");
        }
        GeneralPath o = new GeneralPath(0);
        int[] array = new int[10];
        int n2 = 0;
        if (i < n) {
            do {
                array[n2++] = this.hitToCaret(TextHitInfo.leading(i));
                int charLTR = this.textLine.isCharLTR(i);
                while (++i < n && (this.textLine.isCharLTR(i) ? 1 : 0) == charLTR) {}
                array[n2++] = this.hitToCaret(TextHitInfo.trailing(i - 1));
                if (n2 == array.length) {
                    int[] array2 = new int[array.length + 10];
                    System.arraycopy(array, 0, array2, 0, n2);
                    array = array2;
                }
                else {
                    continue;
                }
            } while (i < n);
        }
        else {
            n2 = 2;
            array[0] = (array[1] = this.hitToCaret(TextHitInfo.leading(i)));
        }
        for (int j = 0; j < n2; j += 2) {
            o.append(this.caretBoundingShape(array[j], array[j + 1], rectangle2D), false);
        }
        if (i != n) {
            if ((this.textLine.isDirectionLTR() && i == 0) || (!this.textLine.isDirectionLTR() && n == this.characterCount)) {
                GeneralPath leftShape = this.leftShape(rectangle2D);
                if (!leftShape.getBounds().isEmpty()) {
                    o.append(leftShape, false);
                }
            }
            if ((this.textLine.isDirectionLTR() && n == this.characterCount) || (!this.textLine.isDirectionLTR() && i == 0)) {
                GeneralPath rightShape = this.rightShape(rectangle2D);
                if (!rightShape.getBounds().isEmpty()) {
                    o.append(rightShape, false);
                }
            }
        }
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            o = (GeneralPath)layoutPath.mapShape(o);
            return o;
        }
        return o;
    }
    
    public Shape getLogicalHighlightShape(int n, int n2) {
        return this.getLogicalHighlightShape(n, n2, this.getNaturalBounds());
    }
    
    public Shape getBlackBoxBounds(int i, int n) {
        this.ensureCache();
        if (i > n) {
            i = (n = n);
        }
        if (i < 0 || n > this.characterCount) {
            throw new IllegalArgumentException("Invalid range passed to TextLayout.getBlackBoxBounds()");
        }
        GeneralPath pSrc = new GeneralPath(1);
        if (i < this.characterCount) {
            Rectangle2D charBounds;
            for (; i < n; ++i) {
                charBounds = this.textLine.getCharBounds(i);
                if (!charBounds.isEmpty()) {
                    pSrc.append(charBounds, false);
                }
            }
        }
        if (TextLayout.dx != 0.0f || TextLayout.dy != 0.0f) {
            pSrc = (GeneralPath)AffineTransform.getTranslateInstance((double)TextLayout.dx, (double)TextLayout.dy).createTransformedShape(pSrc);
        }
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            pSrc = (GeneralPath)layoutPath.mapShape(pSrc);
            return pSrc;
        }
        return pSrc;
    }
    
    private float caretToPointDistance(float[] array, float n, float n2) {
        n = (this.isVerticalLine ? n2 : n);
        n2 = (this.isVerticalLine ? (-n) : n2);
        return n - array[0] + n2 * array[1];
    }
    
    public TextHitInfo hitTestChar(float x, float y, Rectangle2D rectangle2D) {
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            Point2D.Float float1 = new Point2D.Float(x, y);
            layoutPath.pointToPath(float1, float1);
            x = float1.x;
            y = float1.y;
        }
        if (this.isVertical()) {
            if ((double)y < rectangle2D.getMinY()) {
                return TextHitInfo.leading(0);
            }
            if ((double)y >= rectangle2D.getMaxY()) {
                return TextHitInfo.trailing(this.characterCount - 1);
            }
        }
        else {
            if ((double)x < rectangle2D.getMinX()) {
                return this.isLeftToRight() ? TextHitInfo.leading(0) : TextHitInfo.trailing(this.characterCount - 1);
            }
            if ((double)x >= rectangle2D.getMaxX()) {
                return this.isLeftToRight() ? TextHitInfo.trailing(this.characterCount - 1) : TextHitInfo.leading(0);
            }
        }
        double n = Double.MAX_VALUE;
        int n2 = 0;
        int characterCount = -1;
        CoreMetrics coreMetrics = null;
        float n3 = 0.0f;
        float n4 = 0.0f;
        float italicAngle = 0.0f;
        float n5 = 0.0f;
        float n6 = 0.0f;
        float n7 = 0.0f;
        for (int i = 0; i < this.characterCount; ++i) {
            if (this.textLine.caretAtOffsetIsValid(i)) {
                if (characterCount == -1) {
                    characterCount = i;
                }
                CoreMetrics coreMetrics2 = this.textLine.getCoreMetricsAt(i);
                if (coreMetrics2 != coreMetrics) {
                    coreMetrics = coreMetrics2;
                    float n8;
                    if (coreMetrics2.baselineIndex == -1) {
                        n8 = -(this.textLine.getMetrics().ascent - coreMetrics2.ascent) + coreMetrics2.ssOffset;
                    }
                    else if (coreMetrics2.baselineIndex == -2) {
                        n8 = this.textLine.getMetrics().descent - coreMetrics2.descent + coreMetrics2.ssOffset;
                    }
                    else {
                        n8 = coreMetrics2.effectiveBaselineOffset(this.baselineOffsets) + coreMetrics2.ssOffset;
                    }
                    float n9 = (coreMetrics2.descent - coreMetrics2.ascent) / 2.0f - n8;
                    n6 = n9 * coreMetrics2.italicAngle;
                    n5 = n8 + n9;
                    n7 = (n5 - y) * (n5 - y);
                }
                float n10 = this.textLine.getCharXPosition(i) + (this.textLine.getCharAdvance(i) / 2.0f - n6);
                double sqrt = Math.sqrt((double)(4.0f * (n10 - x) * (n10 - x) + n7));
                if (sqrt < n) {
                    n = sqrt;
                    n2 = i;
                    characterCount = -1;
                    n3 = n10;
                    n4 = n5;
                    italicAngle = coreMetrics2.italicAngle;
                }
            }
        }
        int n11 = this.textLine.isCharLTR(n2) == x < n3 - (y - n4) * italicAngle;
        if (characterCount == -1) {
            characterCount = this.characterCount;
        }
        return n11 ? TextHitInfo.leading(n2) : TextHitInfo.trailing(characterCount - 1);
    }
    
    public TextHitInfo hitTestChar(float n, float n2) {
        return this.hitTestChar(n, n2, this.getNaturalBounds());
    }
    
    @Override
    public int hashCode() {
        if (this.hashCodeCache == 0) {
            this.ensureCache();
            this.hashCodeCache = this.textLine.hashCode();
        }
        return this.hashCodeCache;
    }
    
    @Override
    public boolean equals(Object o) {
        return o instanceof TextLayout && this.equals((TextLayout)o);
    }
    
    public boolean equals(TextLayout textLayout) {
        if (textLayout == null) {
            return false;
        }
        if (textLayout == this) {
            return true;
        }
        this.ensureCache();
        return this.textLine.equals(textLayout.textLine);
    }
    
    @Override
    public String toString() {
        this.ensureCache();
        return this.textLine.toString();
    }
    
    public void draw(Graphics2D graphics2D, float n, float n2) {
        if (graphics2D == null) {
            throw new IllegalArgumentException("Null Graphics2D passed to TextLayout.draw()");
        }
        this.textLine.draw(graphics2D, n - TextLayout.dx, n2 - TextLayout.dy);
        return;
    }
    
    TextLine getTextLineForTesting() {
        return this.textLine;
    }
    
    private static int sameBaselineUpTo(Font font, char[] array, int n, int n2) {
        return n2;
    }
    
    static byte getBaselineFromGraphic(GraphicAttribute graphicAttribute) {
        int n = (byte)graphicAttribute.getAlignment();
        if (n == -2 || n == -1) {
            return 0;
        }
        return n;
    }
    
    public Shape getOutline(AffineTransform affineTransform) {
        this.ensureCache();
        Shape shape = this.textLine.getOutline(affineTransform);
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            shape = layoutPath.mapShape(shape);
            return shape;
        }
        return shape;
    }
    
    public LayoutPath getLayoutPath() {
        return this.textLine.getLayoutPath();
    }
    
    public void hitToPoint(TextHitInfo textHitInfo, Point2D point2D) {
        if (textHitInfo == null || point2D == null) {
            throw new NullPointerException(((textHitInfo == null) ? "hit" : "point") + " can't be null");
        }
        this.ensureCache();
        this.checkTextHit(textHitInfo);
        float charYPosition = 0.0f;
        int charIndex = textHitInfo.getCharIndex();
        int leadingEdge = textHitInfo.isLeadingEdge();
        int n;
        float charLinePosition;
        if (charIndex == -1 || charIndex == this.textLine.characterCount()) {
            n = this.textLine.isDirectionLTR();
            charLinePosition = ((n == ((charIndex == -1) ? 1 : 0)) ? 0.0f : this.lineMetrics.advance);
        }
        else {
            n = this.textLine.isCharLTR(charIndex);
            charLinePosition = this.textLine.getCharLinePosition(charIndex, (boolean)leadingEdge);
            charYPosition = this.textLine.getCharYPosition(charIndex);
        }
        point2D.setLocation((double)charLinePosition, (double)charYPosition);
        LayoutPathImpl layoutPath = this.textLine.getLayoutPath();
        if (layoutPath != null) {
            layoutPath.pathToPoint(point2D, n != leadingEdge, point2D);
        }
        return;
    }
    
    static TextHitInfo access$000(TextLayout textLayout, TextHitInfo textHitInfo, TextHitInfo textHitInfo2) {
        return textLayout.getStrongHit(textHitInfo, textHitInfo2);
    }
    
    static {
        DEFAULT_CARET_POLICY = new TextLayout.CaretPolicy();
    }
    
    public static class CaretPolicy
    {
        public TextHitInfo getStrongCaret(TextHitInfo textHitInfo, TextHitInfo textHitInfo2, TextLayout textLayout) {
            return TextLayout.access$000(textLayout, textHitInfo, textHitInfo2);
        }
    }
}
